/*
# X-BASED HEXTRIS
#
#  Hextris.c
#
###
#
#  Copyright (c) 1993 - 2006	David Albert Bagley, bagleyd@tux.org
#
#  Taken from GENERIC X-BASED TETRIS
#
#  Copyright (c) 1992 - 1995	Q. Alex Zhao, azhao@cc.gatech.edu
#
#			All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Methods file for Hextris */

#include "file.h"
#include "rngs.h"
#include "sound.h"
#include "HextrisP.h"

static Point hexagon_unit[MAX_ORIENT][6] = {
	{ {0, 0}, {2, 0}, {1, 1}, {-1, 1}, {-2, 0}, {-1, -1} },
	{ {0, 0}, {1, 1}, {0, 2}, {-1, 1}, {-1, -1}, {0, -2} }
};
static Point hexagon_list[MAX_ORIENT][7];
static Point hexagon_border[7];

#ifdef WINVER
#ifndef LOGPATH
#define LOGPATH "/usr/tmp"
#endif
#ifndef INIFILE
#define INIFILE "whextris.ini"
#endif

#define SECTION "setup"
#else

#ifndef LOGPATH
#ifdef VMS
#define LOGPATH "SYS$SCRATCH:"
#else
#define LOGPATH "/usr/tmp"
#endif
#endif

static void ResizeGame(HextrisWidget w);
static void SizeGame(HextrisWidget w);
static Boolean SetValuesGame(Widget current, Widget request, Widget renew);
static void DestroyGame(Widget old);
static void QuitGame(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void InitializeGame(Widget request, Widget renew);
static void ExposeGame(Widget renew, XEvent *event, Region region);
static void HideGame(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void ClearGame(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void EnterGame(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void LeaveGame(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void ReflectGame(HextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void RotateGame(HextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void MoveGameLeft(HextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void MoveGameRight(HextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void DropGame(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void MoveGameInput(HextrisWidget w, int direction);
static void IncrementGameLevel(HextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void DecrementGameLevel(HextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void IncrementGameUnits(HextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void DecrementGameUnits(HextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void IncrementGameColumns(HextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void DecrementGameColumns(HextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void FixedFillGame(HextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void SuddenAppearGame(HextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void CornersGame(HextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void MixedGame(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void BonusGame(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void CwGame(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void PlainPieceGame(HextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void ShowNextGame(HextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void GridGame(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void SoundGame(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void PauseGame(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill0Game(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill1Game(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill2Game(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill3Game(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill4Game(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill5Game(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill6Game(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill7Game(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill8Game(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill9Game(HextrisWidget w, XEvent *event, char **args, int nArgs);
static void Fill10Game(HextrisWidget w, XEvent *event, char **args, int nArgs);

static char defaultTranslationsGame[] =
"<KeyPress>q: Quit()\n\
 Ctrl<KeyPress>C: Quit()\n\
 <KeyPress>osfCancel: Hide()\n\
 <KeyPress>Escape: Hide()\n\
 <KeyPress>osfEscape: Hide()\n\
 Ctrl<KeyPress>[: Hide()\n\
 <KeyPress>0x1B: Hide()\n\
 <Btn4Down>: Reflect()\n\
 <KeyPress>i: Reflect()\n\
 <KeyPress>e: Reflect()\n\
 <KeyPress>KP_Divide: Reflect()\n\
 <KeyPress>R5: Reflect()\n\
 <Btn2Down>: Rotate()\n\
 <KeyPress>k: Rotate()\n\
 <KeyPress>d: Rotate()\n\
 <KeyPress>Up: Rotate()\n\
 <KeyPress>osfUp: Rotate()\n\
 <KeyPress>KP_Up: Rotate()\n\
 <KeyPress>R8: Rotate()\n\
 <KeyPress>KP_8: Rotate()\n\
 <KeyPress>Begin: Rotate()\n\
 <KeyPress>KP_5: Rotate()\n\
 <KeyPress>R11: Rotate()\n\
 <Btn1Down>: MoveLeft()\n\
 <KeyPress>j: MoveLeft()\n\
 <KeyPress>s: MoveLeft()\n\
 <KeyPress>Left: MoveLeft()\n\
 <KeyPress>osfLeft: MoveLeft()\n\
 <KeyPress>KP_Left: MoveLeft()\n\
 <KeyPress>KP_4: MoveLeft()\n\
 <KeyPress>R10: MoveLeft()\n\
 <Btn3Down>: MoveRight()\n\
 <KeyPress>l: MoveRight()\n\
 <KeyPress>f: MoveRight()\n\
 <KeyPress>Right: MoveRight()\n\
 <KeyPress>osfRight: MoveRight()\n\
 <KeyPress>KP_Right: MoveRight()\n\
 <KeyPress>KP_6: MoveRight()\n\
 <KeyPress>R12: MoveRight()\n\
 <Btn5Down>: Drop()\n\
 <KeyPress>Return: Drop()\n\
 <KeyPress>KP_Enter: Drop()\n\
 <KeyPress>Down: Drop()\n\
 <KeyPress>osfDown: Drop()\n\
 <KeyPress>KP_Down: Drop()\n\
 <KeyPress>KP_2: Drop()\n\
 <KeyPress>R14: Drop()\n\
 <KeyPress>KP_Space: Drop()\n\
 <KeyPress>space: Drop()\n\
 <Btn4Down>: Rotate()\n\
 <Btn5Down>: Drop()\n\
 <KeyPress>c: Clear()\n\
 <KeyPress>>: IncrementLevel()\n\
 <KeyPress>.: IncrementLevel()\n\
 <KeyPress><: DecrementLevel()\n\
 <KeyPress>+: IncrementUnits()\n\
 <KeyPress>=: IncrementUnits()\n\
 <KeyPress>-: DecrementUnits()\n\
 <KeyPress>_: DecrementUnits()\n\
 <KeyPress>]: IncrementColumns()\n\
 <KeyPress>}: IncrementColumns()\n\
 <KeyPress>[: DecrementColumns()\n\
 <KeyPress>{: DecrementColumns()\n\
 <KeyPress>x: FixedFill()\n\
 <KeyPress>r: SuddenAppear()\n\
 <KeyPress>o: Corners()\n\
 <KeyPress>m: Mixed()\n\
 <KeyPress>b: Bonus()\n\
 <KeyPress>w: Cw()\n\
 <KeyPress>p: PlainPiece()\n\
 <KeyPress>n: ShowNext()\n\
 Shift<KeyPress>3: Grid()\n\
 Shift<KeyPress>2: Sound()\n\
 <KeyPress>|: Pause()\n\
 <KeyPress>`: Fill0()\n\
 <KeyPress>1: Fill1()\n\
 <KeyPress>2: Fill2()\n\
 <KeyPress>3: Fill3()\n\
 <KeyPress>4: Fill4()\n\
 <KeyPress>5: Fill5()\n\
 <KeyPress>6: Fill6()\n\
 <KeyPress>7: Fill7()\n\
 <KeyPress>8: Fill8()\n\
 <KeyPress>9: Fill9()\n\
 <KeyPress>0: Fill10()\n\
 <EnterWindow>: Enter()\n\
 <LeaveWindow>: Leave()";

/*
 <KeyPress>,: DecrementLevel()\n\
#if defined(UP_REFLECT)
 <KeyPress>KP_Up: Reflect()\n\
 <KeyPress>R8: Reflect()\n\
 <KeyPress>KP_8: Reflect()\n\
#endif
*/

static XtActionsRec actionsListGame[] =
{
	{(char *) "Quit", (XtActionProc) QuitGame},
	{(char *) "Hide", (XtActionProc) HideGame},
	{(char *) "Reflect", (XtActionProc) ReflectGame},
	{(char *) "Rotate", (XtActionProc) RotateGame},
	{(char *) "MoveLeft", (XtActionProc) MoveGameLeft},
	{(char *) "MoveRight", (XtActionProc) MoveGameRight},
	{(char *) "Drop", (XtActionProc) DropGame},
	{(char *) "Clear", (XtActionProc) ClearGame},
	{(char *) "IncrementLevel", (XtActionProc) IncrementGameLevel},
	{(char *) "DecrementLevel", (XtActionProc) DecrementGameLevel},
	{(char *) "IncrementUnits", (XtActionProc) IncrementGameUnits},
	{(char *) "DecrementUnits", (XtActionProc) DecrementGameUnits},
	{(char *) "IncrementColumns", (XtActionProc) IncrementGameColumns},
	{(char *) "DecrementColumns", (XtActionProc) DecrementGameColumns},
	{(char *) "FixedFill", (XtActionProc) FixedFillGame},
	{(char *) "SuddenAppear", (XtActionProc) SuddenAppearGame},
	{(char *) "Corners", (XtActionProc) CornersGame},
	{(char *) "Mixed", (XtActionProc) MixedGame},
	{(char *) "Bonus", (XtActionProc) BonusGame},
	{(char *) "Cw", (XtActionProc) CwGame},
	{(char *) "PlainPiece", (XtActionProc) PlainPieceGame},
	{(char *) "ShowNext", (XtActionProc) ShowNextGame},
	{(char *) "Grid", (XtActionProc) GridGame},
	{(char *) "Sound", (XtActionProc) SoundGame},
	{(char *) "Pause", (XtActionProc) PauseGame},
	{(char *) "Fill0", (XtActionProc) Fill0Game},
	{(char *) "Fill1", (XtActionProc) Fill1Game},
	{(char *) "Fill2", (XtActionProc) Fill2Game},
	{(char *) "Fill3", (XtActionProc) Fill3Game},
	{(char *) "Fill4", (XtActionProc) Fill4Game},
	{(char *) "Fill5", (XtActionProc) Fill5Game},
	{(char *) "Fill6", (XtActionProc) Fill6Game},
	{(char *) "Fill7", (XtActionProc) Fill7Game},
	{(char *) "Fill8", (XtActionProc) Fill8Game},
	{(char *) "Fill9", (XtActionProc) Fill9Game},
	{(char *) "Fill10", (XtActionProc) Fill10Game},
	{(char *) "Enter", (XtActionProc) EnterGame},
	{(char *) "Leave", (XtActionProc) LeaveGame}
};

static XtResource resourcesGame[] =
{
	/* 30*10+2,24*10+2 */
	{XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
	 XtOffset(HextrisWidget, core.width),
	 XtRString, (caddr_t) "314"},
	{XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	 XtOffset(HextrisWidget, core.height),
	 XtRString, (caddr_t) "434"},
	{XtNmono, XtCMono, XtRBoolean, sizeof (Boolean),
	 XtOffset(HextrisWidget, hextris.mono),
	 XtRString, (caddr_t) "FALSE"},
	{XtNreverse, XtCReverse, XtRBoolean, sizeof (Boolean),
	 XtOffset(HextrisWidget, hextris.reverse),
	 XtRString, (caddr_t) "FALSE"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(HextrisWidget, hextris.foreground),
	 XtRString, (caddr_t) XtDefaultForeground},
	{XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(HextrisWidget, hextris.background),
	 XtRString, (caddr_t) XtDefaultBackground},
	{XtNframeColor, XtCColor, XtRPixel, sizeof (Pixel),
	 XtOffset(HextrisWidget, hextris.frameColor),
	 XtRString, (caddr_t) "cyan" /*XtDefaultForeground*/},
	{XtNborderColor, XtCColor, XtRPixel, sizeof (Pixel),
	 XtOffset(HextrisWidget, hextris.borderColor),
	 XtRString, (caddr_t) "gray25" /*XtDefaultForeground*/},
	{XtNdelay, XtCDelay, XtRInt, sizeof (int),
	 XtOffset(HextrisWidget, hextris.delay),
	 XtRString, (caddr_t) "1000"},
	{XtNsound, XtCSound, XtRBoolean, sizeof (Boolean),
	 XtOffset(HextrisWidget, hextris.sound),
	 XtRString, (caddr_t) "FALSE"},
	{XtNbumpSound, XtCBumpSound, XtRString, sizeof (String),
	 XtOffset(HextrisWidget, hextris.bumpSound),
	 XtRString, (caddr_t) BUMPSOUND},
	{XtNcolumns, XtCColumns, XtRInt, sizeof (int),
	 XtOffset(HextrisWidget, hextris.columns),
	 XtRString, (caddr_t) "10"},
	{XtNlevel, XtCLevel, XtRInt, sizeof (int),
	 XtOffset(HextrisWidget, hextris.level),
	 XtRString, (caddr_t) "3"},
	{XtNunits, XtCUnits, XtRInt, sizeof (int),
	 XtOffset(HextrisWidget, hextris.units),
	 XtRString, (caddr_t) "4"},
	{XtNfill, XtCFill, XtRInt, sizeof (int),
	 XtOffset(HextrisWidget, hextris.fill),
	 XtRString, (caddr_t) "0"},
	{XtNfixedFill, XtCFixedFill, XtRBoolean, sizeof (Boolean),
	 XtOffset(HextrisWidget, hextris.fixedFill),
	 XtRString, (caddr_t) "TRUE"},
	{XtNsuddenAppear, XtCSuddenAppear, XtRBoolean, sizeof (Boolean),
	 XtOffset(HextrisWidget, hextris.suddenAppear),
	 XtRString, (caddr_t) "FALSE"},
	{XtNcorners, XtCCorners, XtRBoolean, sizeof (Boolean),
	 XtOffset(HextrisWidget, hextris.corners),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmixed, XtCMixed, XtRBoolean, sizeof (Boolean),
	 XtOffset(HextrisWidget, hextris.mixed),
	 XtRString, (caddr_t) "FALSE"},
	{XtNbonus, XtCBonus, XtRBoolean, sizeof (Boolean),
	 XtOffset(HextrisWidget, hextris.bonus),
	 XtRString, (caddr_t) "TRUE"},
	{XtNcw, XtCCw, XtRBoolean, sizeof (Boolean),
	 XtOffset(HextrisWidget, hextris.cw),
	 XtRString, (caddr_t) "TRUE"},
	{XtNplainPiece, XtCPlainPiece, XtRBoolean, sizeof (Boolean),
	 XtOffset(HextrisWidget, hextris.plainPiece),
	 XtRString, (caddr_t) "FALSE"},
	{XtNshowNext, XtCShowNext, XtRBoolean, sizeof (Boolean),
	 XtOffset(HextrisWidget, hextris.showNext),
	 XtRString, (caddr_t) "TRUE"},
	{XtNgrid, XtCGrid, XtRBoolean, sizeof (Boolean),
	 XtOffset(HextrisWidget, hextris.grid),
	 XtRString, (caddr_t) "TRUE"},
	{XtNscore, XtCBoolean, XtRInt, sizeof (int),
	 XtOffset(HextrisWidget, hextris.score),
	 XtRString, (caddr_t) "0"},
	{XtNcompletedRows, XtCBoolean, XtRInt, sizeof (int),
	 XtOffset(HextrisWidget, hextris.completedRows),
	 XtRString, (caddr_t) "0"},
	{XtNuserName, XtCUserName, XtRString, sizeof (String),
	 XtOffset(HextrisWidget, hextris.userName),
	 XtRString, (caddr_t) ""},
	{XtNscoreFile, XtCScoreFile, XtRString, sizeof (String),
	 XtOffset(HextrisWidget, hextris.scoreFile),
	 XtRString, (caddr_t) ""},
	{XtNscoreOnly, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(HextrisWidget, hextris.scoreOnly),
	 XtRString, (caddr_t) "FALSE"},
	{XtNversionOnly, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(HextrisWidget, hextris.versionOnly),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmenu, XtCMenu, XtRInt, sizeof (int),
	 XtOffset(HextrisWidget, hextris.menu),
	 XtRString, (caddr_t) "-1"},
	{XtNstart, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(HextrisWidget, hextris.started),
	 XtRString, (caddr_t) "FALSE"},
	{XtNact, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(HextrisWidget, hextris.act),
	 XtRString, (caddr_t) "FALSE"},
	{XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
	 XtOffset(HextrisWidget, hextris.select),
	 XtRCallback, (caddr_t) NULL}
};

HextrisClassRec hextrisClassRec =
{
	{
		(WidgetClass) & widgetClassRec,		/* superclass */
		(char *) "Hextris",	/* class name */
		sizeof (HextrisRec),	/* widget size */
		NULL,		/* class initialize */
		NULL,		/* class part initialize */
		FALSE,		/* class inited */
		(XtInitProc) InitializeGame,	/* initialize */
		NULL,		/* initialize hook */
		XtInheritRealize,	/* realize */
		actionsListGame,	/* actions */
		XtNumber(actionsListGame),	/* num actions */
		resourcesGame,	/* resources */
		XtNumber(resourcesGame),	/* num resources */
		NULLQUARK,	/* xrm class */
		TRUE,		/* compress motion */
		TRUE,		/* compress exposure */
		TRUE,		/* compress enterleave */
		TRUE,		/* visible interest */
		(XtWidgetProc) DestroyGame,	/* destroy */
		(XtWidgetProc) ResizeGame,	/* resize */
		(XtExposeProc) ExposeGame,	/* expose */
		(XtSetValuesFunc) SetValuesGame,	/* set values */
		NULL,		/* set values hook */
		XtInheritSetValuesAlmost,	/* set values almost */
		NULL,		/* get values hook */
		NULL,		/* accept focus */
		XtVersion,	/* version */
		NULL,		/* callback private */
		defaultTranslationsGame,	/* tm table */
		NULL,		/* query geometry */
		NULL,		/* display accelerator */
		NULL		/* extension */
	},
	{
		0		/* ignore */
	}
};

WidgetClass hextrisWidgetClass = (WidgetClass) & hextrisClassRec;

static void
setGame(HextrisWidget w, int reason)
{
	hextrisCallbackStruct cb;

	cb.reason = reason;
	XtCallCallbacks((Widget) w, (char *) XtNselectCallback, &cb);
}

static void
setGameNext(HextrisWidget w, int reason,
		int randomNumber, int randomRotation, int randomReflection,
		int units, Boolean corners, Boolean mixed, Boolean newBonus)
{
	hextrisCallbackStruct cb;

	cb.reason = reason;
	cb.randomNumber = randomNumber;
	cb.randomRotation = randomRotation;
	cb.randomReflection = randomReflection;
	cb.units = units;
	cb.corners = corners;
	cb.mixed = mixed;
	cb.newBonus = newBonus;
	XtCallCallbacks((Widget) w, (char *) XtNselectCallback, &cb);
}

#endif

#include "image.h"

static Pixmap dr = 0; /* dummy for future double buffering */

static const int delay[MAX_LEVELS] = {
	100, 92, 84, 75, 68, 60, 52, 45, 39, 32,
	26, 21, 16, 12, 8, 5, 3, 2, 1, 0
};

static void
CheckThings(HextrisWidget w)
{
	char *buf1 = NULL, *buf2 = NULL;

	if (w->hextris.units < MIN_UNITS || w->hextris.units > MAX_UNITS) {
		intCat(&buf1, "Number of units out of bounds, use ",
			MIN_UNITS);
		stringCat(&buf2, buf1, "...");
		free(buf1);
		intCat(&buf1, buf2, MAX_UNITS);
		free(buf2);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULT_UNITS);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
		w->hextris.units = DEFAULT_UNITS;
		w->hextris.next.units = w->hextris.units;
	}
	if (w->hextris.columns < MIN_COLUMNS) {
		intCat(&buf1, "Number of columns out of bounds, use at least ",
			MIN_COLUMNS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULTCOLUMNS);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
		w->hextris.columns = DEFAULTCOLUMNS;
	}
	if (!LEGAL(w->hextris.units, w->hextris.corners,
			w->hextris.mixed, FALSE)) {
		intCat(&buf1, "Illegal configuration of units ",
			w->hextris.units);
		stringCat(&buf2, buf1, ", corners ");
		free(buf1);
		intCat(&buf1, buf2, (int) w->hextris.corners);
		free(buf2);
		stringCat(&buf2, buf1, ", mixed ");
		free(buf1);
		intCat(&buf1, buf2, (int) w->hextris.mixed);
		free(buf2);
		stringCat(&buf2, buf1, ", defaulting to units ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULT_UNITS);
		free(buf2);
		stringCat(&buf2, buf1, ", corners ");
		free(buf1);
		intCat(&buf1, buf2, (int) DEFAULTCORNERS);
		free(buf2);
		stringCat(&buf2, buf1, ", mixed ");
		free(buf1);
		intCat(&buf1, buf2, (int) DEFAULTMIXED);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
		w->hextris.units = DEFAULT_UNITS;
		w->hextris.corners = FALSE;
		w->hextris.mixed = FALSE;
		w->hextris.next.units = w->hextris.units;
		w->hextris.next.corners = w->hextris.corners;
		w->hextris.next.mixed = w->hextris.mixed;
	}
}

#define FACTOR 0.7

#ifdef WINVER
#define MAXINTENSITY 0xFF
int
brighter(const int light)
{
	int i = (int) ((1 - FACTOR) * MAXINTENSITY);
	int temp = light;

	if (temp < i)
		temp = i;
	return MIN(temp / FACTOR, MAXINTENSITY);
}

int
darker(const int light)
{
	return (int) (light * FACTOR);
}

#else
#define MAXINTENSITY 0xFFFF

Pixel
brighter(HextrisWidget w, Pixel pixel)
{
	Display *display = XtDisplay(w);
	Colormap colormap = DefaultColormap(display, DefaultScreen(display));

	XColor color;
	int i = (int) ((1 - FACTOR) * MAXINTENSITY);

	color.pixel = pixel;
	VOID XQueryColor(display, colormap, &color);
	if (color.red < i)
		color.red = i;
	if (color.green < i)
		color.green = i;
	if (color.blue < i)
		color.blue = i;
	color.red = (unsigned short) MIN(color.red / FACTOR, MAXINTENSITY);
	color.green = (unsigned short) MIN(color.green / FACTOR, MAXINTENSITY);
	color.blue = (unsigned short) MIN(color.blue / FACTOR, MAXINTENSITY);
	if (XAllocColor(display, colormap, &color))
		return color.pixel;
	(void) fprintf(stderr,
		"Hextris: Could not allocate brighter color %ld: %d %d %d\n",
		color.pixel, color.red, color.green, color.blue);
	return pixel;
}

Pixel
darker(HextrisWidget w, Pixel pixel)
{
	Display *display = XtDisplay(w);
	Colormap colormap = DefaultColormap(display, DefaultScreen(display));
	XColor color;

	color.pixel = pixel;
	VOID XQueryColor(display, colormap, &color);
	color.red = (unsigned short) (color.red * FACTOR);
	color.green = (unsigned short) (color.green * FACTOR);
	color.blue = (unsigned short) (color.blue * FACTOR);
	if (XAllocColor(display, colormap, &color))
		return color.pixel;
	(void) fprintf(stderr,
		"Hextris: Could not allocate darker color %ld: %d %d %d\n",
		color.pixel, color.red, color.green, color.blue);
	return pixel;
}
#endif

#if 0
static void
drawHex(Window win, GC gc, int i, int j)
{
	if (j < 0)
		return;
	hexagon_list[current.corners][0].x = i * w->hextris.hexWidth +
		(((j + w->hextris.invert) % 2) ?
		(w->hextris.hexWidth + 1) / 2 : w->hextris.hexWidth + 1) +
		((current.corners) ? 0 : -w->hextris.hexWidth / 4);
	hexagon_list[current.corners][0].y = j * w->hextris.hexHeight +
		((current.corners) ? 2 : w->hextris.hexHeight / 4 - 1);
	VOID XFillPolygon(display, win, gc, hexagon_list[current.corners],
		MAX_SIDES, Convex, CoordModePrevious);
}
#endif

static void
drawUnitBorder(HextrisWidget w, int i, int j)
{
	GC gc;

	if (w->hextris.grid)
		gc = w->hextris.borderGC;
	else
		gc = w->hextris.inverseGC;
	hexagon_border[0].x = i * w->hextris.hexWidth + w->hextris.offset.x +
		(((j + w->hextris.invert) % 2) ?
		(w->hextris.hexWidth + 1) / 2 : w->hextris.hexWidth + 1);
	hexagon_border[0].y = j * w->hextris.hexHeight +
		w->hextris.offset.y;
	POLYLINE(w, dr, gc,
		hexagon_border, MAX_SIDES + 1, False);
}

#define EDGE 2

void
fillPolygon(HextrisWidget w, int pm, int color, Point *relPoints,
		Boolean corners, Boolean graphicalXor)
{
	Point midPoint, tempList[MAX_SIDES + 1], innerPoints[MAX_SIDES + 1];
       	Point pts[5];
	int i;
	GC gc = w->hextris.thingGC[color];
	GC darkGC = w->hextris.thingDarkGC[color];
	GC brightGC = w->hextris.thingBrightGC[color];
	GC darkerGC = w->hextris.thingDarkerGC[color];
	GC brighterGC = w->hextris.thingBrighterGC[color];
	GC *shade[MAX_SIDES];

#ifdef WINVER
	/* simulate Xor */
	if (graphicalXor) {
		gc = RGB(255 - GetRValue(gc),
			255 - GetGValue(gc), 255 - GetBValue(gc));
		darkGC = RGB(255 - GetRValue(darkGC),
			255 - GetGValue(darkGC), 255 - GetBValue(darkGC));
		brightGC = RGB(255 - GetRValue(brightGC),
			255 - GetGValue(brightGC), 255 - GetBValue(brightGC));
		darkerGC = RGB(255 - GetRValue(darkerGC),
			255 - GetGValue(darkerGC), 255 - GetBValue(darkerGC));
		brighterGC = RGB(255 - GetRValue(brighterGC),
			255 - GetGValue(brighterGC),
			255 - GetBValue(brighterGC));
	}
#endif

	POLYGON(w, dr, gc, gc, relPoints, MAX_SIDES, True, False);
	if (!w->hextris.plainPiece) {
		tempList[0] = relPoints[0];
		for (i = 1; i <= MAX_SIDES; i++) {
			tempList[i].x = tempList[i - 1].x + relPoints[i].x;
			tempList[i].y = tempList[i - 1].y + relPoints[i].y;
		}
		midPoint.x = (tempList[0].x + tempList[MAX_SIDES / 2].x) / 2;
		midPoint.y = (tempList[0].y + tempList[MAX_SIDES / 2].y) / 2;
		if (!corners) {
			for (i = 0; i < MAX_SIDES; i++) {
				tempList[i].x = (tempList[i].x +
					tempList[i + 1].x) / 2;
				tempList[i].y = (tempList[i].y +
					tempList[i + 1].y) / 2;
			}
			tempList[MAX_SIDES].x = tempList[0].x;
			tempList[MAX_SIDES].y = tempList[0].y;
		}
		for (i = 0; i <= MAX_SIDES; i++) {
			innerPoints[i].x = ((EDGE - 1) * tempList[i].x +
				midPoint.x) / EDGE;
			innerPoints[i].y = ((EDGE - 1) * tempList[i].y +
				midPoint.y) / EDGE;
		}
		POLYGON(w, dr, gc, gc, innerPoints, MAX_SIDES, True, True);
		for (i = 0; i < MAX_SIDES; i++) {
			if ((pm & (1 << (MAX_SIDES - 1 - i))) != 0) {
				pts[0] = tempList[(i + MAX_SIDES - 1) %
					MAX_SIDES];
				pts[1] = tempList[i];
				pts[2] = innerPoints[i];
				pts[3] = innerPoints[(i + MAX_SIDES - 1) %
					MAX_SIDES];
				pts[4] = tempList[(i + MAX_SIDES - 1) %
					MAX_SIDES];
				POLYGON(w, dr, darkGC, brightGC, pts, 4,
					True, True);
			}
		}
		return;
	}
	shade[0] = &brighterGC;
	shade[1] = &brightGC;
	shade[2] = &darkGC;
	shade[3] = &darkerGC;
	shade[4] = &darkGC;
	shade[5] = &brightGC;
	tempList[0] = relPoints[0];
	for (i = 1; i <= MAX_SIDES; i++) {
		tempList[i].x = tempList[i - 1].x + relPoints[i].x;
		tempList[i].y = tempList[i - 1].y + relPoints[i].y;
	}
	midPoint.x = (tempList[0].x + tempList[MAX_SIDES / 2].x) / 2;
	midPoint.y = (tempList[0].y + tempList[MAX_SIDES / 2].y) / 2;
	for (i = 0; i <= MAX_SIDES; i++) {
		innerPoints[i].x = (tempList[i].x + midPoint.x) / 2;
		innerPoints[i].y = (tempList[i].y + midPoint.y) / 2;
	}
	POLYGON(w, dr, gc, gc, innerPoints, MAX_SIDES, True, True);
	for (i = 0; i < MAX_SIDES; i++) {
		pts[0] = tempList[(i + MAX_SIDES - 1) % MAX_SIDES];
		pts[1] = tempList[i];
		pts[2] = innerPoints[i];
		pts[3] = innerPoints[(i + MAX_SIDES - 1) % MAX_SIDES];
		pts[4] = tempList[(i + MAX_SIDES - 1) % MAX_SIDES];
		POLYGON(w, dr, *(shade[i]), *(shade[i]), pts, 4, True, True);
	}
}

static void
drawUnit(HextrisWidget w, int pmid, int cid, int i, int j,
		Boolean graphicalXor)
{
	Boolean corners = w->hextris.current.corners;

	if (i >= 0 && i < w->hextris.columns &&
			j >= 0 && j < w->hextris.rows) {
#if 0
		hexagon_list[corn][0].x = i * w->hextris.hexWidth +
			(((j + w->hextris.invert) % 2) ? (w->hextris.hexWidth + 1) / 2 : w->hextris.hexWidth + 1) +
			((corners) ? 0 : -w->hextris.hexWidth / 4);
		hexagon_list[corn][0].y = j * w->hextris.hexHeight +
			((corners) ? 2 : w->hextris.hexHeight / 4 - 1);
#endif
		hexagon_border[0].x = i * w->hextris.hexWidth +
			w->hextris.offset.x +
			(((j + w->hextris.invert) % 2) ? (w->hextris.hexWidth +
			1) / 2 : w->hextris.hexWidth + 1);
		hexagon_border[0].y = j * w->hextris.hexHeight +
			w->hextris.offset.y + ((corners) ? 0 : -1);
		if (corners) {
			fillPolygon(w, pmid, cid, hexagon_border,
				True, graphicalXor);
		} else {
			Point hexagon_borderTrunc[MAX_SIDES + 1];
			Point aPoint, bPoint, aPointTrunc, bPointTrunc;
			int k;

			aPoint.x = aPoint.y = 0;
			bPoint.x = bPoint.y = 0;
			aPointTrunc.x = aPointTrunc.y = 0;
			bPointTrunc.x = -hexagon_border[1].x / 2;
			bPointTrunc.y = hexagon_border[1].y / 2;
			for (k = 0; k < MAX_SIDES; k++) {
				aPoint.x += hexagon_border[k + 1].x;
				aPoint.y += hexagon_border[k + 1].y;
				aPointTrunc.x = (aPoint.x + bPoint.x) / 2;
				aPointTrunc.y = (aPoint.y + bPoint.y) / 2;
				hexagon_borderTrunc[k + 1].x =
					aPointTrunc.x - bPointTrunc.x;
				hexagon_borderTrunc[k + 1].y =
					aPointTrunc.y - bPointTrunc.y;
				bPoint = aPoint;
				bPointTrunc = aPointTrunc;
			}
			hexagon_borderTrunc[0].x = hexagon_border[0].x -
				hexagon_border[1].x / 2;
			hexagon_borderTrunc[0].y = hexagon_border[0].y +
				hexagon_border[1].y / 2 + 1;
			fillPolygon(w, pmid, cid, hexagon_borderTrunc,
				False, graphicalXor);
		}
	}
}

static void
clearUnit(HextrisWidget w, int i, int j)
{
	if (i >= 0 && i < w->hextris.columns && j >= 0 && j < w->hextris.rows) {
		hexagon_border[0].x = i * w->hextris.hexWidth +
			w->hextris.offset.x +
			(((j + w->hextris.invert) % 2) ?
			(w->hextris.hexWidth + 1) / 2 :
			w->hextris.hexWidth + 1);
		hexagon_border[0].y = j * w->hextris.hexHeight +
			w->hextris.offset.y + ((w->hextris.corners) ? 0 : -1);
		POLYGON(w, dr, w->hextris.inverseGC, w->hextris.inverseGC,
			hexagon_border, MAX_SIDES,
			True, False);
	}
}

static void
xorUnit(HextrisWidget w, int i, int j, int k)
{
	hexagon_border[0].x = i * w->hextris.hexWidth +
		w->hextris.offset.x +
		(((j + w->hextris.invert) % 2) ?
		(w->hextris.hexWidth + 1) / 2 : w->hextris.hexWidth + 1);
	hexagon_border[0].y = j * w->hextris.hexHeight +
		w->hextris.offset.y + ((w->hextris.corners) ? 0 : -1);
#ifdef WINVER
	if (w->hextris.field[j * w->hextris.columns + i].pmid >= 0) {
		drawUnit(w,
			w->hextris.field[j * w->hextris.columns + i].pmid,
			w->hextris.field[j * w->hextris.columns + i].cid,
			i, j, (k % 2 == 0));
	}
#else
	POLYGON(w, dr, w->hextris.xorGC, w->hextris.xorGC,
		hexagon_border, MAX_SIDES,
		True, False);
#endif
}

static void
moveUnit(HextrisWidget w, int i, int j)
{
	if (w->hextris.field[j * w->hextris.columns + i].pmid == -1) {
		xorUnit(w, i, j, 0);
	} else {
		drawUnit(w,
			w->hextris.field[j * w->hextris.columns + i].pmid,
			w->hextris.field[j * w->hextris.columns + i].cid,
			i, j, False);
	}
	drawUnitBorder(w, i, j);
}

static void
drawPartialField(HextrisWidget w, int start, int finish)
{
	int i, j, offset;

	for (j = start; j < finish; j++) {
		offset = j * w->hextris.columns;
		for (i = 0; i < w->hextris.columns; i++) {
			if (w->hextris.field[offset + i].pmid >= 0) {
				drawUnit(w, w->hextris.field[offset + i].pmid,
					w->hextris.field[offset + i].cid,
					i, j, False);
				drawUnitBorder(w, i, j);
			}
		}
	}
}

static void
drawField(HextrisWidget w)
{
	drawPartialField(w, 0, w->hextris.rows);
}

static void
clearField(HextrisWidget w)
{
	int i, j;

	for (j = 0; j < w->hextris.rows; j++) {
		for (i = 0; i < w->hextris.columns; i++) {
			clearUnit(w, i, j);
		}
	}
}

static void
drawThing(HextrisWidget w)
{
	int i, j;

	for (i = 0; i < HEX_ROW_COL(w->hextris.curThing.size); i++)
		for (j = 0; j < HEX_ROW_COL(w->hextris.curThing.size); j++) {
			int yj = w->hextris.curThing.ypos + j;
			int xi = w->hextris.curThing.xpos + i +
				XPOS(w->hextris.curThing.size, w->hextris.curThing.ypos, j, w->hextris.invert);

			if (yj >= 0 && inHexagon(w->hextris.curThing.size, i, j) != 0 &&
					tris[w->hextris.curThing.units].polyhex
					[w->hextris.curThing.polyhexNumber].shape[j][i] != 0) {
				drawUnit(w, polyPixmap(i, j,
					(tris[w->hextris.curThing.units].polyhex
					[w->hextris.curThing.polyhexNumber].shape),
					w->hextris.curThing.size),
					w->hextris.curThing.colorNumber,
					xi, yj, False);
				drawUnitBorder(w, xi, yj);
			}
		}
}

static void
drawThingDiff(HextrisWidget w, Thing oldThing)
{
	int i, j;
	int s = HEX_ROW_COL(w->hextris.curThing.size);

	for (i = 0; i < s; i++)
		for (j = 0; j < s; j++) {
			int yj = w->hextris.curThing.ypos + j;
			int xi = w->hextris.curThing.xpos + i +
				XPOS(w->hextris.curThing.size, w->hextris.curThing.ypos, j, w->hextris.invert);

			if (yj >= 0 && inHexagon(w->hextris.curThing.size, i, j) != 0 &&
					tris[w->hextris.curThing.units].polyhex
					[w->hextris.curThing.polyhexNumber].shape[j][i] != 0) {
				drawUnit(w, polyPixmap(i, j,
					(tris[w->hextris.curThing.units].polyhex
					[w->hextris.curThing.polyhexNumber].shape),
w->hextris.curThing.size),
					w->hextris.curThing.colorNumber,
					xi, yj, False);
				drawUnitBorder(w, xi, yj);
			}
		}
	for (i = 0; i < s; i++) {
		int xi = oldThing.xpos + i;

		if (xi >= -1 && xi <= w->hextris.columns)
			for (j = 0; j < s; j++) {
				int yj = oldThing.ypos + j;
				int xi_temp = xi;
				int oy = yj - w->hextris.curThing.ypos;
				int ox = xi - w->hextris.curThing.xpos;

				if ((w->hextris.curThing.ypos - oldThing.ypos) % 2)
					ox += (((w->hextris.curThing.size + j) % 2) ?
						(((oldThing.ypos + w->hextris.invert) % 2) ? -1 : 0) :
						(((oldThing.ypos + !w->hextris.invert) % 2) ? 1 : 0));
				xi_temp += XPOS(w->hextris.curThing.size, oldThing.ypos, j, w->hextris.invert);
				if (yj >= 0 && yj < w->hextris.rows &&
					xi_temp >= 0 && xi_temp < w->hextris.columns &&
					(tris[oldThing.units].polyhex[oldThing.polyhexNumber].shape[j][i] != 0) &&
					((ox < 0) || (ox >= s) ||
					(oy < 0) || (oy >= s) ||
					tris[w->hextris.curThing.units].polyhex
					[w->hextris.curThing.polyhexNumber].shape[oy][ox] == 0)) {
					clearUnit(w, xi_temp, yj);
					drawUnitBorder(w, xi_temp, yj);
				}
			}
	}
}

static void
drawGrid(HextrisWidget w)
{
	int i, j;

	for (j = 0; j < w->hextris.rows; j++)
		for (i = 0; i < w->hextris.columns; i++)
			drawUnitBorder(w, i, j);
}

static void
EraseFrame(HextrisWidget w)
{
	FILLRECTANGLE(w, dr, w->hextris.inverseGC,
		0, 0, w->core.width, w->core.height);
}

static void
DrawFrame(HextrisWidget w, Boolean focus)
{
	DRAWRECTANGLE(w, dr,
		(focus) ? w->hextris.frameGC : w->hextris.borderGC,
		w->hextris.offset.x - 1, w->hextris.offset.y - 2,
		w->hextris.viewWidth + 1, w->hextris.viewHeight + 1);
}

static void
drawNext(HextrisWidget w)
{
	setGameNext(w, GAME_DRAWNEXT,
		w->hextris.nextThing.randomNumber,
		w->hextris.nextThing.randomRotation,
		w->hextris.nextThing.randomReflection,
		w->hextris.units, w->hextris.corners, w->hextris.mixed,
		w->hextris.next.bonus);
}

static void
drawStatus(HextrisWidget w)
{
	setGame(w, GAME_STATUS);
}

static void
redoNext(HextrisWidget w)
{
	int corners = (w->hextris.next.corners) ? 1 : 0;
	int turnStyle = tris[w->hextris.next.units -
		MIN_UNITS].mode[corners].turnStyle;
	int next_start, i;

	w->hextris.nextThing.units = w->hextris.next.units - MIN_UNITS +
		w->hextris.nextThing.bonus;
	next_start = tris[w->hextris.nextThing.units].number
		[(w->hextris.nextThing.bonus) ?
		NOMIX : w->hextris.next.mixed][turnStyle];
	if (!next_start) {
		DISPLAY_ERROR("Error, no polyhexes for this combination");
	}
	next_start = w->hextris.nextThing.randomNumber % next_start;
	w->hextris.nextThing.colorNumber = next_start;

	if (w->hextris.next.mixed && !w->hextris.nextThing.bonus) {
		w->hextris.nextThing.units = 0;
		while (next_start >= tris[w->hextris.nextThing.units].number
				[NOMIX][turnStyle]) {
			next_start -= tris[w->hextris.nextThing.units].number
				[NOMIX][turnStyle];
			w->hextris.nextThing.units++;
		}
	}
	w->hextris.nextThing.polyhexNumber =
		tris[w->hextris.nextThing.units].mode
		[corners].start[next_start][turnStyle];
	if (tris[w->hextris.nextThing.units].mode[corners].turnStyle > NONE)
		for (i = 0; i < w->hextris.nextThing.randomRotation; i++)
			w->hextris.nextThing.polyhexNumber =
				tris[w->hextris.nextThing.units].polyhex
				[w->hextris.nextThing.polyhexNumber].rotation;
	if (tris[w->hextris.nextThing.units].mode[corners].turnStyle == ALL)
		for (i = 0; i < w->hextris.nextThing.randomReflection; i++)
			w->hextris.nextThing.polyhexNumber =
				tris[w->hextris.nextThing.units].polyhex
				[w->hextris.nextThing.polyhexNumber].reflection;
	w->hextris.nextThing.size =tris[w->hextris.nextThing.units].polyhex
			[w->hextris.nextThing.polyhexNumber].size;
	/* w->hextris.nextThing.xpos = NRAND(w->hextris.columns - w->hextris.nextThing.size + 1); */
	w->hextris.nextThing.xpos =
		(w->hextris.columns - w->hextris.nextThing.size) / 2;
	w->hextris.nextThing.ypos = -tris[w->hextris.nextThing.units].polyhex
		[w->hextris.nextThing.polyhexNumber].startHeight[(w->hextris.suddenAppear) ? JUMPIN : GRADUAL];
	w->hextris.nextThing.direction = (NRAND(2) % 2) ? 1 : -1;
}

static void
initFill(HextrisWidget w)
{
	int i, j, col, offset, fixedUnits;

	if (w->hextris.fixedFill) {
		/* This algorithm randomly picks fixedUnits per row
		 * (4 for DEFAULTCOLUMNS, ie. 10/2-1)
		 * This is more in keeping with the original */
		fixedUnits = w->hextris.columns / 2 - 1;
		for (j = 0; j < w->hextris.rows; j++) {
			offset = j * w->hextris.columns;
			for (i = 0; i < w->hextris.columns; i++) {
				w->hextris.field[offset + i].pmid = -1;
				w->hextris.field[offset + i].cid = 0;
			}
			if (j >= w->hextris.rows - w->hextris.fill)
				for (i = 0; i < fixedUnits; i++) {
					do
						col = NRAND(w->hextris.columns);
					while (w->hextris.field[offset + col].pmid != -1);
					w->hextris.field[offset + col].pmid = BITMAPS - 1;
					w->hextris.field[offset + col].cid = NRAND(w->hextris.numberPolyhexes);
				}
		}
	} else {
		/* This algorithm randomly fills boxes to a certain level
		 * The problem with this is that it could potentially fill
		 * up a whole row */
		for (j = 0; j < w->hextris.rows; j++) {
			offset = j * w->hextris.columns;
			for (i = 0; i < w->hextris.columns; i++)
				if ((j >= w->hextris.rows - w->hextris.fill) &&
						(NRAND(2) == 0)) {
					w->hextris.field[offset + i].pmid = BITMAPS - 1;
					w->hextris.field[offset + i].cid = NRAND(w->hextris.numberPolyhexes);
				} else {
					w->hextris.field[offset + i].pmid = -1;
					w->hextris.field[offset + i].cid = 0;
				}
		}
	}
}

static void
fillLines(HextrisWidget w)
{
	int i, j, k;
	int corners = (w->hextris.current.corners) ? 1 : 0;
	int mixed = (w->hextris.current.mixed) ? 1 : 0;
	int turnStyle = tris[w->hextris.current.units -
		MIN_UNITS].mode[corners].turnStyle;

	FLUSH(w);
	for (j = 0; j <= w->hextris.rows / 2; j++) {
		k = w->hextris.rows - j - 1;
		for (i = 0; i < w->hextris.columns; i++) {
			drawUnit(w, BITMAPS - 1, NRAND(tris[w->hextris.current.units -
				MIN_UNITS].number[mixed][turnStyle]), i, j,
				False);
			drawUnitBorder(w, i, j);
			drawUnit(w, BITMAPS - 1, NRAND(tris[w->hextris.current.units -
				MIN_UNITS].number[mixed][turnStyle]), i, k,
				False);
			drawUnitBorder(w, i, k);
		}
		FLUSH(w);
		Sleep(100);
	}
	Sleep(1000);
}

static Boolean
overlapping(HextrisWidget w)
{
	int i, j;
	int x = w->hextris.curThing.xpos, y = w->hextris.curThing.ypos;

	for (i = 0; i < HEX_ROW_COL(w->hextris.curThing.size); i++)
		for (j = 0; j < HEX_ROW_COL(w->hextris.curThing.size); j++)
			if (tris[w->hextris.curThing.units].polyhex
					[w->hextris.curThing.polyhexNumber].shape[j][i] != 0) {
				int xi = x + i + XPOS(w->hextris.curThing.size,
					y, j, w->hextris.invert);
				int yj = y + j;

				if ((yj >= w->hextris.rows) || (xi < 0) ||
						(xi >= w->hextris.columns))
					return True;
				if (w->hextris.suddenAppear) { /* creationalism */
/* This method does not allow turning polyhex to an area above screen */
					if ((yj < 0) || (w->hextris.field[yj *
							w->hextris.columns + xi].pmid >= 0))
						return True;
				} else { /* evolution */
/* This method one can turn polyhex to an area above of screen,
 * also part of the polyhex may not be visible initially */
					if ((yj >= 0) && (w->hextris.field[yj *
						w->hextris.columns + xi].pmid >= 0))
					return True;
				}
			}
	return False;
}

/* only used for corners, walls are soft */
static Boolean
overlapped(HextrisWidget w, int dir, int fall)
{
	Boolean blocked[2];
	int i, j, k, l;
	int x = w->hextris.curThing.xpos, y = w->hextris.curThing.ypos;
	int oldy = w->hextris.curThing.ypos - ((fall) ? 1 : 0);
	int oldx = w->hextris.curThing.xpos -
		((fall) ? XINC(y, dir, w->hextris.invert) : dir);

	blocked[0] = blocked[1] = False;
	for (i = 0; i < HEX_ROW_COL(w->hextris.curThing.size); i++)
		for (j = 0; j < HEX_ROW_COL(w->hextris.curThing.size); j++) {
			int xi = x + i;
			int yj = y + j;
			int oldyj = oldy + j;

			if ((yj >= 0) && tris[w->hextris.curThing.units].polyhex
						[w->hextris.curThing.polyhexNumber].shape[j][i] != 0) {
				if (fall) {
					k = oldx + i + XPOS(w->hextris.curThing.size, oldy, j, w->hextris.invert) + dir;
					if (w->hextris.suddenAppear) { /* creationalism */
/* This method there is a wall on top of screen */
						if ((oldyj < 0) ||
#ifdef HARD_SIDE
								(k < 0) || (k >= w->hextris.columns) ||
								(w->hextris.field[oldyj * w->hextris.columns + k].pmid >= 0)
#else
								((k >= 0) && (k < w->hextris.columns) &&
								(w->hextris.field[oldyj * w->hextris.columns + k].pmid >= 0))
#endif
								)
							blocked[0] = True;
					} else { /* evolution */
/* This method there is no wall on top of screen */
						if ((oldyj >= 0) &&
#ifdef HARD_SIDE
								((k < 0) || (k >= w->hextris.columns) ||
								(w->hextris.field[oldyj * w->hextris.columns + k].pmid >= 0))
#else
								(k >= 0) && (k < w->hextris.columns) &&
								(w->hextris.field[oldyj * w->hextris.columns + k].pmid >= 0)
#endif
								)
							blocked[0] = True;
					}
					k = xi + XPOS(w->hextris.curThing.size, y, j, w->hextris.invert) - dir;
					if (
#ifdef HARD_SIDE
							(k < 0) || (k >= w->hextris.columns) ||
							(w->hextris.field[(y + j) * w->hextris.columns + k].pmid >= 0)
#else
							((k >= 0) && (k < w->hextris.columns) &&
							(w->hextris.field[(y + j) * w->hextris.columns + k].pmid >= 0))
#endif
							)
						blocked[1] = True;
				} else {
					l = oldx + i + XINC(y + 1, dir, w->hextris.invert) +
						XPOS(w->hextris.curThing.size, y + 1, j, w->hextris.invert);
					k = y - 1 + j;
					if (w->hextris.suddenAppear) { /* creationalism */
/* This method there is a wall on top of screen */
						if ((k < 0) || (w->hextris.field[k * w->hextris.columns + l].pmid >= 0))
							blocked[0] = True;
					} else { /* evolution */
/* This method there is no wall on top of screen */
						if ((k >= 0) && (w->hextris.field[k * w->hextris.columns + l].pmid >= 0))
							blocked[0] = True;
					}
					k = y + 1 + j;
					if ((k >= w->hextris.rows) || (w->hextris.field[k * w->hextris.columns + l].pmid >= 0))
						blocked[1] = True;
				}
			}
		}
	return blocked[0] && blocked[1];
}

static Boolean
collision(HextrisWidget w, int dir)
{
	int i, j, k;
	int x = w->hextris.curThing.xpos, y = w->hextris.curThing.ypos;
	int newy = w->hextris.curThing.ypos + 1;
	int newx = w->hextris.curThing.xpos + XINC(newy, dir, w->hextris.invert);

	for (i = 0; i < HEX_ROW_COL(w->hextris.curThing.size); i++)
		for (j = 0; j < HEX_ROW_COL(w->hextris.curThing.size); j++)
			if ((y + j >= -1) && tris[w->hextris.curThing.units].polyhex
					[w->hextris.curThing.polyhexNumber].shape[j][i] != 0) {
				k = x + i + XINC(newy, dir, w->hextris.invert) +
					XPOS(w->hextris.curThing.size, newy, j, w->hextris.invert);
				if ((y + j >= w->hextris.rows - 1) ||
						(k < 0) || (k >= w->hextris.columns) ||
						(w->hextris.field[(newy + j) * w->hextris.columns + k].pmid >= 0))
					return True;
			}
	if (w->hextris.current.corners) {
		Boolean blocked[2];

		blocked[0] = blocked[1] = False;
		for (i = 0; i < HEX_ROW_COL(w->hextris.curThing.size); i++)
			for (j = 0; j < HEX_ROW_COL(w->hextris.curThing.size); j++) {
				int yj = y + j;

				if ((yj >= -1) && tris[w->hextris.curThing.units].polyhex
							[w->hextris.curThing.polyhexNumber].shape[j][i] != 0) {
					k = x + i + XPOS(w->hextris.curThing.size, y, j, w->hextris.invert) + dir;
					if (w->hextris.suddenAppear) { /* creationalism */
/* This method does not allow a turning polyhex to an area above screen */
						if ((yj < 0) ||
#ifdef HARD_SIDE
								(k < 0) || (k >= w->hextris.columns) ||
								(w->hextris.field[yj * w->hextris.columns + k].pmid >= 0)
#else
								((k >= 0) && (k < w->hextris.columns) &&
								(w->hextris.field[yj * w->hextris.columns + k].pmid >= 0))
#endif
								)
							blocked[0] = True;
					} else { /* evolution */
/* This method one can turn polyhex to an area above of screen,
 * also part of the polyhex may not be visible initially */
						if ((yj >= 0) &&
#ifdef HARD_SIDE
								((k < 0) || (k >= w->hextris.columns) ||
								(w->hextris.field[yj * w->hextris.columns + k].pmid >= 0))
#else
								(k >= 0) && (k < w->hextris.columns) &&
								(w->hextris.field[yj * w->hextris.columns + k].pmid >= 0)
#endif
								)
							blocked[0] = True;
					}
					k = newx + i + XPOS(w->hextris.curThing.size, newy, j, w->hextris.invert) - dir;
					if (
#ifdef HARD_SIDE
							(k < 0) || (k >= w->hextris.columns) ||
							(w->hextris.field[(newy + j) * w->hextris.columns + k].pmid >= 0)
#else
							((k >= 0) && (k < w->hextris.columns) &&
							(w->hextris.field[(newy + j) * w->hextris.columns + k].pmid >= 0))
#endif
							)
						blocked[1] = True;
				}
			}
		return blocked[0] && blocked[1];
	}
	return False;
}

static Boolean
atBottom(HextrisWidget w)
{
	if (collision(w, w->hextris.curThing.direction))
		return collision(w, -w->hextris.curThing.direction);
	return False;
}

static void
putHex(HextrisWidget w)
{
	int i, j;
	int x = w->hextris.curThing.xpos;

	for (i = 0; i < HEX_ROW_COL(w->hextris.curThing.size); i++)
		for (j = 0; j < HEX_ROW_COL(w->hextris.curThing.size); j++) {
			int yj = w->hextris.curThing.ypos + j;
			int xi = x + i + XPOS(w->hextris.curThing.size,
				w->hextris.curThing.ypos, j, w->hextris.invert);

			if ((yj >= 0) && tris[w->hextris.curThing.units].polyhex
					[w->hextris.curThing.polyhexNumber].shape[j][i] != 0) {
				w->hextris.field[yj * w->hextris.columns + xi].pmid = (polyPixmap(i, j,
					(tris[w->hextris.curThing.units].polyhex
					[w->hextris.curThing.polyhexNumber].shape),
					w->hextris.curThing.size)) % BITMAPS;
				w->hextris.field[yj * w->hextris.columns + xi].cid =
					w->hextris.curThing.colorNumber;
			}
		}
}

static int
checkLines(HextrisWidget w)
{
	int *lSet, nSet = 0;
	int i, j, k, y, offset;

	if (!(lSet = (int *) malloc(sizeof (int) * w->hextris.rows))) {
		DISPLAY_ERROR("Not enough memory (lSet), exiting.");
	}
	/* Find filled rows */
	for (j = 0; j < w->hextris.rows; j++) {
		lSet[j] = 0;
		offset = j * w->hextris.columns;
		for (i = 0; i < w->hextris.columns; i++)
			if (w->hextris.field[offset + i].pmid >= 0)
				lSet[j]++;
		if (lSet[j] == w->hextris.columns)
			nSet++;
	}

	if (nSet == 0) {
		free(lSet);
		return 0;
	}
	for (k = 0; k < 2 * NUM_FLASHES; k++) {
		for (j = 0; j < w->hextris.rows; j++) {
			if (lSet[j] == w->hextris.columns) {
				for (i = 0; i < w->hextris.columns; i++) {
					xorUnit(w, i, j, k);
					drawUnitBorder(w, i, j);
				}
			}
		}
		FLUSH(w);
		Sleep(100);
	}
	/* Shrink */
	for (j = w->hextris.rows - 1; j >= 0; j--)
		if (lSet[j] == w->hextris.columns) {
			for (y = j; y > 0; y--)
				for (i = 0; i < w->hextris.columns; i++) {
					w->hextris.field[y * w->hextris.columns + i] =
						w->hextris.field[(y - 1) * w->hextris.columns + i];
					moveUnit(w, i, y);
				}
			for (i = 0; i < w->hextris.columns; i++) {
				w->hextris.field[i].pmid = -1;
				clearUnit(w, i, 0);
				drawUnitBorder(w, i, 0);
			}
			/* drawPartialField(w, 0, j + 1); */
			for (i = j; i > 0; i--)
				lSet[i] = lSet[i - 1];
			lSet[0] = 0;
			if (j > 0)
				for (i = 0; i < w->hextris.columns; i++) {
					int tmp = w->hextris.field[j * w->hextris.columns + i].pmid;

					if ((tmp >= 0) && ((tmp != CHECKLEFT_DOWN(tmp)) ||
							(tmp != CHECKRIGHT_DOWN(tmp)))) {
						w->hextris.field[j * w->hextris.columns + i].pmid = CHECKRIGHT_DOWN(CHECKLEFT_DOWN(tmp));
						drawUnit(w, w->hextris.field[j * w->hextris.columns + i].pmid,
							w->hextris.field[j * w->hextris.columns + i].cid, i, j, False);
					}
				}
			FLUSH(w);
			j++;
			for (k = j; k < w->hextris.rows; k++) {
				offset = k * w->hextris.columns;
				for (i = 0; i < w->hextris.columns; i++) {
					int tmp = w->hextris.field[offset + i].pmid;

					if (tmp >= 0) {
						tmp = CHECKLEFT_DOWN(tmp);
						tmp = CHECKRIGHT_DOWN(tmp);
						tmp = CHECKLEFT_UP(tmp);
						tmp = CHECKRIGHT_UP(tmp);
						w->hextris.field[offset + i].pmid = tmp;
					}
				}
			}
#ifdef USE_SOUND
			if (w->hextris.sound) {
				playSound((char *) BUMPSOUND);
			}
#endif
		}
	free(lSet);
	/*clearField(w);*/
	EraseFrame(w);
	if (nSet % 2 == 1) {
		w->hextris.invert = !w->hextris.invert;
	}
	DrawFrame(w, w->hextris.focus);
	if (w->hextris.grid)
		drawGrid(w);
	drawField(w);
	FLUSH(w);
	return nSet;
}

static void
newThing(HextrisWidget w)
{
	w->hextris.curThing = w->hextris.nextThing;
	/*w->hextris.current = w->hextris.next;*/
	w->hextris.current.units = w->hextris.next.units;
	w->hextris.current.corners = w->hextris.next.corners;
	w->hextris.current.mixed = w->hextris.next.mixed;
	w->hextris.current.bonus = w->hextris.next.bonus;
	w->hextris.nextThing.randomNumber = LRAND();
	w->hextris.nextThing.randomRotation = NRAND(MAX_SIDES);
	w->hextris.nextThing.randomReflection = NRAND(2);
	w->hextris.nextThing.bonus = (w->hextris.next.bonus) ? 1 : 0;
	redoNext(w);
}

static void
ClearThings(HextrisWidget w, Boolean draw)
{
	if (draw)
		clearField(w);
	initFill(w);
	if (draw) {
		if (w->hextris.grid)
			drawGrid(w);
		drawField(w);
	}
	w->hextris.delay = delay[w->hextris.level - 1];
	w->hextris.score = w->hextris.fill * w->hextris.level * 10;
	w->hextris.started = True;
	w->hextris.paused = False;
	w->hextris.pauseStart = 0;
	w->hextris.completedRows = 0;
	w->hextris.newRows = 0;
	drawStatus(w);
}

static void
gameOver(HextrisWidget w, Boolean fast)
{
	if (!w->hextris.started)
		return;
	if (!fast)
		fillLines(w);
	setGame(w, GAME_GAMEOVER);
	ClearThings(w, True);
}

static void
tryMove(HextrisWidget w, move_t move)
{
	int corners = (w->hextris.current.corners) ? 1 : 0;
	Thing old, mid;
	int i;

	old = w->hextris.curThing;
	switch (move) {
	case FALL:
		w->hextris.curThing.ypos++;
		w->hextris.curThing.xpos += XINC(w->hextris.curThing.ypos, w->hextris.curThing.direction, w->hextris.invert);
		if (overlapping(w) || (corners &&
				overlapped(w, w->hextris.curThing.direction, True))) {
			w->hextris.curThing = old;
			w->hextris.curThing.direction = -w->hextris.curThing.direction;
			w->hextris.curThing.ypos++;
			w->hextris.curThing.xpos += XINC(w->hextris.curThing.ypos, w->hextris.curThing.direction, w->hextris.invert);
		}
		if (overlapping(w) || (corners &&
				overlapped(w, w->hextris.curThing.direction, True)))
			w->hextris.curThing = old;
		else
			drawThingDiff(w, old);
		break;
	case DROP:
		do {
			do {
				mid = w->hextris.curThing;
				w->hextris.curThing.ypos++;
				w->hextris.curThing.xpos +=
					XINC(w->hextris.curThing.ypos,
					w->hextris.curThing.direction,
					w->hextris.invert);
				w->hextris.score += w->hextris.level + w->hextris.fill;
			} while (!overlapping(w) && !(corners &&
					overlapped(w, w->hextris.curThing.direction, True)));
			w->hextris.curThing = mid;
			w->hextris.curThing.direction = -w->hextris.curThing.direction;
			w->hextris.curThing.ypos++;
			w->hextris.curThing.xpos += XINC(w->hextris.curThing.ypos, w->hextris.curThing.direction, w->hextris.invert);
		} while (!overlapping(w) && !(corners &&
			overlapped(w, w->hextris.curThing.direction, True)));
		w->hextris.curThing = mid;
		drawThingDiff(w, old);
		break;
	case MOVE_LEFT:
		w->hextris.curThing.xpos--;
		if (overlapping(w) || (corners && overlapped(w, -1, False)))
			w->hextris.curThing = old;
		else
			drawThingDiff(w, old);
		break;
	case MOVE_RIGHT:
		w->hextris.curThing.xpos++;
		if (overlapping(w) || (corners && overlapped(w, 1, False)))
			w->hextris.curThing = old;
		else
			drawThingDiff(w, old);
		break;
	case ROTATE:
		if (tris[w->hextris.current.units - MIN_UNITS].mode[corners].turnStyle > NONE) {
			if (w->hextris.cw)
				for (i = 0; i < MAX_SIDES - 1; i++)
					w->hextris.curThing.polyhexNumber = tris[w->hextris.curThing.units].polyhex
						[w->hextris.curThing.polyhexNumber].rotation;
			else /* ccw */
				w->hextris.curThing.polyhexNumber = tris[w->hextris.curThing.units].polyhex
					[w->hextris.curThing.polyhexNumber].rotation;
			w->hextris.curThing.xpos = old.xpos;
			w->hextris.curThing.ypos = old.ypos;
		}
		if (!overlapping(w))
			drawThingDiff(w, old);
		else
			w->hextris.curThing = old;
		break;

	case REFLECT: /* reflect on y axis */
		if (tris[w->hextris.current.units - MIN_UNITS].mode[corners].turnStyle == ALL) {
			w->hextris.curThing.polyhexNumber = tris[w->hextris.curThing.units].polyhex
				[w->hextris.curThing.polyhexNumber].reflection;
			w->hextris.curThing.xpos = old.xpos;
			w->hextris.curThing.ypos = old.ypos;
		}
		if (!overlapping(w))
			drawThingDiff(w, old);
		else
			w->hextris.curThing = old;
		break;
	}
}

static Boolean
moveOne(HextrisWidget w, move_t move)
{
	int lines;

	if ((move == DROP) || ((move == FALL) && atBottom(w))) {
		tryMove(w, move);
		putHex(w);
		lines = checkLines(w);
		w->hextris.score += (w->hextris.fill + w->hextris.level +
			lines + w->hextris.suddenAppear) *
			(w->hextris.current.bonus ? 1 : 3) *
			(w->hextris.showNext ? 1 : 3) /
			(1 + w->hextris.current.mixed);
		w->hextris.score += (w->hextris.showNext ? 3 : 5) *
			w->hextris.level * w->hextris.level * lines * lines *
			(w->hextris.current.bonus ? 3 : 5) *
			w->hextris.startLevel *
			(w->hextris.current.units + w->hextris.current.corners) *
			(w->hextris.current.units + w->hextris.current.corners) /
			(16 * (1 + w->hextris.current.mixed));
		w->hextris.completedRows += lines;
		w->hextris.newRows += lines;
		w->hextris.next.bonus = False;
		if ((w->hextris.newRows > THRESHOLD(w->hextris.level)) &&
				(w->hextris.level < MAX_LEVELS - 1)) {
			w->hextris.level++;
			w->hextris.newRows = 0;
			if (LEGAL(w->hextris.next.units,
					w->hextris.next.corners,
					w->hextris.next.mixed,
					w->hextris.bonus))
				w->hextris.next.bonus = True;
		}
		drawStatus(w);
		if (w->hextris.current.corners != w->hextris.next.corners) {
			EraseFrame(w);
			w->hextris.current = w->hextris.next;
			DrawFrame(w, w->hextris.focus);
			if (w->hextris.grid)
				drawGrid(w);
			drawField(w);
		}
		newThing(w);
		drawThing(w);
		drawNext(w);
		FLUSH(w);
		if (overlapping(w))
			gameOver(w, False);
		return True;
	} else {
		tryMove(w, move);
		if ((w->hextris.newRows > THRESHOLD(w->hextris.level)) &&
				(w->hextris.level < MAX_LEVELS - 1)) {
			w->hextris.next.bonus = False;
			w->hextris.level++;
			w->hextris.newRows = 0;
			if (LEGAL(w->hextris.next.units,
					w->hextris.next.corners,
					w->hextris.next.mixed,
					w->hextris.bonus))
				w->hextris.next.bonus = True;
			drawStatus(w);
			drawNext(w);
		}
		return False;
	}
}

#ifndef WINVER
static
#endif
void
FalldownGame(HextrisWidget w)
{
	if (!w->hextris.paused) {
		w->hextris.gotNew = moveOne(w, FALL) || w->hextris.gotNew;
	} else {
		if (w->hextris.pauseStart >= 32) {
			ClearThings(w, True);
		}
		if (!w->hextris.started)
			w->hextris.pauseStart++;
	}
}

#ifndef WINVER
static
#endif
void
MoveGameInput(HextrisWidget w, int direction)
{
	if (!w->hextris.paused)
		w->hextris.gotNew = moveOne(w, (move_t) direction) ||
			w->hextris.gotNew;
}

static void
ResetThings(HextrisWidget w)
{
	if (w->hextris.field)
		free(w->hextris.field);
	w->hextris.rows = 2 * w->hextris.columns;
	if (!(w->hextris.field = (Field *) malloc(sizeof(Field) *
			w->hextris.rows * w->hextris.columns))) {
		DISPLAY_ERROR("Not enough memory (field), exiting.");
	}
	w->hextris.started = False;
}

#if 0
static void
ResizeThings(HextrisWidget w)
{
}
#endif

#ifdef WINVER
static void
SetValuesGame(HextrisWidget w)
{
	struct tagColor {
		int red, green, blue;
	} color;
	char szBuf[80], buf[20];
	int thing;

	w->hextris.level = GetPrivateProfileInt(SECTION, "level",
		DEFAULTLEVELS, INIFILE);
	w->hextris.units = GetPrivateProfileInt(SECTION, "units",
		DEFAULTUNITS, INIFILE);
	w->hextris.columns = GetPrivateProfileInt(SECTION, "columns",
		DEFAULTCOLUMNS, INIFILE);
	w->hextris.fill = GetPrivateProfileInt(SECTION, "fill",
		0, INIFILE);
	w->hextris.fixedFill = (BOOL) GetPrivateProfileInt(SECTION,
		"fixedFill", TRUE, INIFILE);
	w->hextris.suddenAppear = (BOOL) GetPrivateProfileInt(SECTION,
		"suddenAppear", FALSE, INIFILE);
	w->hextris.corners = (BOOL) GetPrivateProfileInt(SECTION, "corners",
		DEFAULTCORNERS, INIFILE);
	w->hextris.mixed = (BOOL) GetPrivateProfileInt(SECTION, "mixed",
		DEFAULTMIXED, INIFILE);
	w->hextris.bonus = (BOOL) GetPrivateProfileInt(SECTION, "bonus",
		TRUE, INIFILE);
	w->hextris.cw = (BOOL) GetPrivateProfileInt(SECTION, "cw",
		TRUE, INIFILE);
	w->hextris.plainPiece = (BOOL) GetPrivateProfileInt(SECTION,
		"plainPiece", FALSE, INIFILE);
	w->hextris.showNext = (BOOL) GetPrivateProfileInt(SECTION, "showNext",
		FALSE, INIFILE);
	w->hextris.grid = (BOOL) GetPrivateProfileInt(SECTION, "grid",
		TRUE, INIFILE);
	w->hextris.sound = (BOOL) GetPrivateProfileInt(SECTION, "sound",
		FALSE, INIFILE);
	(void) GetPrivateProfileString(SECTION, "bumpSound", BUMPSOUND,
		szBuf, sizeof (szBuf), INIFILE);
	(void) strcpy(w->hextris.bumpSound, szBuf);
	w->hextris.bumpSound[80] = 0;
	w->hextris.mono = (BOOL) GetPrivateProfileInt(SECTION, "mono",
		DEFAULTMONO, INIFILE);
	w->hextris.reverse = (BOOL) GetPrivateProfileInt(SECTION, "reverse",
		DEFAULTREVERSE, INIFILE);
	/* cyan */
	(void) GetPrivateProfileString(SECTION, "frameColor", "0 255 255",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->hextris.frameGC = RGB(color.red, color.green, color.blue);
	/* gray25 */
	(void) GetPrivateProfileString(SECTION, "borderColor", "64 64 64",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->hextris.borderGC = RGB(color.red, color.green, color.blue);
	/* #AEB2C3 */
	(void) GetPrivateProfileString(SECTION, "background", "255 255 255",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->hextris.inverseGC = RGB(color.red, color.green, color.blue);
	for (thing = 0; thing < MAX_START_POLYHEXES; thing++) {
		(void) sprintf(buf, "thingColor%d", thing);
		(void) GetPrivateProfileString(SECTION, buf,
			polyhexColorString[thing],
			szBuf, sizeof (szBuf), INIFILE);
		(void) sscanf(szBuf, "%d %d %d",
			&(color.red), &(color.green), &(color.blue));
		w->hextris.thingGC[thing] =
			RGB(color.red, color.green, color.blue);
		w->hextris.thingBrightGC[thing] =
			RGB(brighter(color.red), brighter(color.green),
			brighter(color.blue));
		w->hextris.thingDarkGC[thing] =
			RGB(darker(color.red), darker(color.green),
			darker(color.blue));
		w->hextris.thingBrighterGC[thing] =
			RGB(brighter(brighter(color.red)),
			brighter(brighter(color.green)),
			brighter(brighter(color.blue)));
		w->hextris.thingDarkerGC[thing] =
			RGB(darker(darker(color.red)),
			darker(darker(color.green)),
			darker(darker(color.blue)));
	}
	(void) GetPrivateProfileString(SECTION, "userName", "Guest",
		szBuf, sizeof (szBuf), INIFILE);
	(void) strcpy(w->hextris.userName, szBuf);
		w->hextris.userName[80] = 0;
	(void) GetPrivateProfileString(SECTION, "scoreFile", "",
		szBuf, sizeof (szBuf), INIFILE);
	(void) strcpy(w->hextris.scoreFile, szBuf);
		w->hextris.scoreFile[80] = 0;
}

void
DestroyGame(HBRUSH brush)
{
	(void) DeleteObject(brush);
	PostQuitMessage(0);
}

#else
static void
GetColor(HextrisWidget w, int thing)
{
	XGCValues values;
	XtGCMask valueMask;
	XColor colorCell, rgb;
	Boolean gotColor = False;

	valueMask = GCForeground | GCBackground;
	if (w->hextris.reverse) {
		values.background = w->hextris.foreground;
	} else {
		values.background = w->hextris.background;
	}
	if (!w->hextris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyhexColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->hextris.thingColor[thing] =
				colorCell.pixel;
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyhexColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyhex ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.background = w->hextris.background;
			values.foreground = w->hextris.foreground;
		}
	}
	if (w->hextris.thingGC[thing])
		XtReleaseGC((Widget) w, w->hextris.thingGC[thing]);
	w->hextris.thingGC[thing] = XtGetGC((Widget) w, valueMask, &values);
	gotColor = False;
	if (!w->hextris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyhexColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->hextris.thingColor[thing] =
				brighter(w, colorCell.pixel);
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyhexColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyhex ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.background = w->hextris.background;
			values.foreground = w->hextris.foreground;
		}
	}
	if (w->hextris.thingBrightGC[thing])
		XtReleaseGC((Widget) w, w->hextris.thingBrightGC[thing]);
	w->hextris.thingBrightGC[thing] = XtGetGC((Widget) w, valueMask,
		&values);
	gotColor = False;
	if (!w->hextris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyhexColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->hextris.thingColor[thing] =
				darker(w, colorCell.pixel);
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyhexColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyhex ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.background = w->hextris.background;
			values.foreground = w->hextris.foreground;
		}
	}
	if (w->hextris.thingDarkGC[thing])
		XtReleaseGC((Widget) w, w->hextris.thingDarkGC[thing]);
	w->hextris.thingDarkGC[thing] = XtGetGC((Widget) w, valueMask,
		&values);
	gotColor = False;
	if (!w->hextris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyhexColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->hextris.thingColor[thing] =
				brighter(w, brighter(w, colorCell.pixel));
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyhexColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyhex ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.background = w->hextris.background;
			values.foreground = w->hextris.foreground;
		}
	}
	if (w->hextris.thingBrighterGC[thing])
		XtReleaseGC((Widget) w, w->hextris.thingBrighterGC[thing]);
	w->hextris.thingBrighterGC[thing] = XtGetGC((Widget) w, valueMask,
		&values);
	gotColor = False;
	if (!w->hextris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyhexColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->hextris.thingColor[thing] =
				darker(w, darker(w, colorCell.pixel));
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyhexColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyhex ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.background = w->hextris.background;
			values.foreground = w->hextris.foreground;
		}
	}
	if (w->hextris.thingDarkerGC[thing])
		XtReleaseGC((Widget) w, w->hextris.thingDarkerGC[thing]);
	w->hextris.thingDarkerGC[thing] = XtGetGC((Widget) w, valueMask, &values);
}

static void
SetAllColors(HextrisWidget w)
{
	XGCValues values;
	XtGCMask valueMask;
	int polyhex;

	valueMask = GCForeground | GCBackground;

	if (w->hextris.reverse) {
		values.background = w->hextris.background;
		values.foreground = w->hextris.foreground;
	} else {
		values.foreground = w->hextris.background;
		values.background = w->hextris.foreground;
	}
	if (w->hextris.inverseGC)
		XtReleaseGC((Widget) w, w->hextris.inverseGC);
	w->hextris.inverseGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->hextris.mono) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.foreground = w->hextris.foreground;
			values.background = w->hextris.background;
		}
	} else {
		values.foreground = w->hextris.frameColor;
		values.background = w->hextris.background;
	}
	if (w->hextris.frameGC)
		XtReleaseGC((Widget) w, w->hextris.frameGC);
	w->hextris.frameGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->hextris.mono) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.foreground = w->hextris.foreground;
			values.background = w->hextris.background;
		}
	} else {
		values.foreground = w->hextris.borderColor;
		values.background = w->hextris.background;
	}
	if (w->hextris.borderGC)
		XtReleaseGC((Widget) w, w->hextris.borderGC);
	w->hextris.borderGC = XtGetGC((Widget) w, valueMask, &values);
	for (polyhex = 0; polyhex < MAX_START_POLYHEXES; polyhex++)
		GetColor(w, polyhex);
	valueMask = GCForeground | GCBackground | GCFunction;
	values.foreground = 0xFFFFFF;
	values.background = 0;
	values.function = GXxor;
	if (w->hextris.xorGC)
		XtReleaseGC((Widget) w, w->hextris.xorGC);
	w->hextris.xorGC = XtGetGC((Widget) w, valueMask, &values);
}

static Boolean
SetValuesGame(Widget current, Widget request, Widget renew)
{
	HextrisWidget c = (HextrisWidget) current, w = (HextrisWidget) renew;
	Boolean redraw = False, setColors = False;

	if (w->hextris.act) {
		FalldownGame(w);
		w->hextris.act = False;
	}
	if (w->hextris.level != c->hextris.level)
		w->hextris.delay = delay[w->hextris.level - 1];
	CheckThings(w);
	if (w->hextris.borderColor != c->hextris.borderColor ||
			w->hextris.frameColor != c->hextris.frameColor ||
			w->hextris.reverse != c->hextris.reverse ||
			w->hextris.mono != c->hextris.mono) {
		SetAllColors(w);
		redraw = True;
	} else if (w->hextris.background != c->hextris.background ||
			w->hextris.foreground != c->hextris.foreground ||
			setColors) {
		SetAllColors(w);
		redraw = True;
	} else if (w->hextris.plainPiece != c->hextris.plainPiece) {
		redraw = True;
	} else if (w->hextris.suddenAppear != c->hextris.suddenAppear) {
		redoNext(w);
	}
	if (w->hextris.columns != c->hextris.columns) {
		w->hextris.paused = True;
		w->hextris.started = False;
		EraseFrame(w);
		SizeGame(w);
#if 0
		ResizeThings(w);
#endif
		newThing(w);
		newThing(w);
		ClearThings(w, True);
		DrawFrame(w, w->hextris.focus);
		if (w->hextris.grid)
			drawGrid(w);
		drawField(w);
		redraw = True;
	} else if (w->hextris.fill != c->hextris.fill) {
		ClearThings(w, True);
		redraw = True;
	} else if (w->hextris.units != c->hextris.units) {
		w->hextris.next.units = w->hextris.units;
		w->hextris.next.bonus = False;
		w->hextris.newRows = 0;
		redoNext(w);
		drawNext(w);
	} else if (w->hextris.corners != c->hextris.corners) {
		w->hextris.next.corners = w->hextris.corners;
		w->hextris.next.bonus = False;
		w->hextris.newRows = 0;
		redoNext(w);
		drawNext(w);
	} else if (w->hextris.mixed != c->hextris.mixed) {
		w->hextris.next.mixed = w->hextris.mixed;
		w->hextris.next.bonus = False;
		w->hextris.newRows = 0;
		redoNext(w);
		drawNext(w);
	} else if (w->hextris.bonus != c->hextris.bonus) {
		w->hextris.next.bonus = False;
		w->hextris.newRows = 0;
		redoNext(w);
		drawNext(w);
	} else if (w->hextris.grid != c->hextris.grid) {
		EraseFrame(w);
		DrawFrame(w, w->hextris.focus);
		if (w->hextris.grid)
			drawGrid(w);
		drawField(w);
	}
	if (w->hextris.menu != -1) {
		switch (w->hextris.menu) {
		case 2:
			w->hextris.menu = -1;
			ClearThings(w, True);
			redraw = True;
			break;
		case 4:
			w->hextris.menu = -1;
			/*ConfirmThings(w);*/
			redraw = True;
			break;
		default:
			w->hextris.menu = -1;
			break;
		}
	}
	return (redraw);
}

static void
QuitGame(HextrisWidget w, XEvent *event, char **args, int nArgs)
{
	Display *display = XtDisplay(w);

	gameOver(w, True);
	XtCloseDisplay(display);
	exit(0);
}

static void
DestroyGame(Widget old)
{
	HextrisWidget w = (HextrisWidget) old;
	int thing;

	for (thing = 0; thing < MAX_START_POLYHEXES; thing++) {
		XtReleaseGC(old, w->hextris.thingGC[thing]);
		XtReleaseGC(old, w->hextris.thingBrightGC[thing]);
		XtReleaseGC(old, w->hextris.thingDarkGC[thing]);
		XtReleaseGC(old, w->hextris.thingBrighterGC[thing]);
		XtReleaseGC(old, w->hextris.thingDarkerGC[thing]);
	}
	XtReleaseGC(old, w->hextris.borderGC);
	XtReleaseGC(old, w->hextris.frameGC);
	XtReleaseGC(old, w->hextris.inverseGC);
	XtReleaseGC(old, w->hextris.xorGC);
	XtRemoveCallbacks(old, XtNselectCallback, w->hextris.select);
}
#endif

#define SQRT2_2 0.70710678 /* sqrt(2) / 2 */

#ifndef WINVER
static
#endif
void
ResizeGame(HextrisWidget w)
{
	int i;
#ifdef WINVER
	RECT rect;

	/* Determine size of client area */
	(void) GetClientRect(w->core.hWnd, &rect);
	w->core.width = rect.right;
	w->core.height = rect.bottom;
#endif

	w->hextris.hexHeight = MIN(3 * (w->core.height - 3) /
		(3 * w->hextris.rows + 1),
		(w->core.width - 2) * 2 * SQRT2_2 /
		(2 * w->hextris.columns + 1));
	w->hextris.hexWidth = (int) (w->hextris.hexHeight / SQRT2_2);
	w->hextris.viewWidth = w->hextris.hexWidth *
		(2 * w->hextris.columns + 1) / 2 + 2;
	w->hextris.viewHeight = w->hextris.hexHeight *
		(3 * w->hextris.rows + 1) / 3 + 2;
	w->hextris.offset.x = ((int) w->core.width -
		w->hextris.viewWidth) / 2;
	w->hextris.offset.y = ((int) w->core.height -
		w->hextris.viewHeight) / 2;

	for (i = 0; i < MAX_SIDES; i++) {
		hexagon_list[CORNERS][i].x =
			(w->hextris.hexWidth - 2) / 2 *
			hexagon_unit[CORNERS][i].x;
		hexagon_list[CORNERS][i].y =
			(2 * w->hextris.hexHeight / 3 - 1) / 2 *
			hexagon_unit[CORNERS][i].y;
		hexagon_list[NOCORN][i].x =
			(w->hextris.hexWidth - 1) / 4 *
			hexagon_unit[NOCORN][i].x;
		hexagon_list[NOCORN][i].y =
			3 * (2 * w->hextris.hexHeight / 3 - 1) / 4 *
			hexagon_unit[NOCORN][i].y;
		hexagon_border[i].x =
			w->hextris.hexWidth / 2 * hexagon_unit[CORNERS][i].x;
		hexagon_border[i].y =
			w->hextris.hexHeight / 3 * hexagon_unit[CORNERS][i].y;
	}
	hexagon_list[CORNERS][MAX_SIDES].x =
		-hexagon_list[CORNERS][MAX_SIDES / 2].x;
	hexagon_list[CORNERS][MAX_SIDES].y =
		-hexagon_list[CORNERS][MAX_SIDES / 2].y;
	hexagon_list[NOCORN][MAX_SIDES].x =
		-hexagon_list[NOCORN][MAX_SIDES / 2].x;
	hexagon_list[NOCORN][MAX_SIDES].y =
		-hexagon_list[NOCORN][MAX_SIDES / 2].y;
	hexagon_border[MAX_SIDES].x = -hexagon_border[MAX_SIDES / 2].x;
	hexagon_border[MAX_SIDES].y = -hexagon_border[MAX_SIDES / 2].y;
}

#ifndef WINVER
static
#endif
void
SizeGame(HextrisWidget w)
{
	ResetThings(w);
	ResizeGame(w);
}

#ifndef WINVER
static
#endif
void
InitializeGame(
#ifdef WINVER
HextrisWidget w, HBRUSH brush
#else
Widget request, Widget renew
#endif
)
{
	int corners, mixed, turnStyle;
#ifndef WINVER
	int poly;
	HextrisWidget w = (HextrisWidget) renew;
#endif

#ifdef WINVER
	SetValuesGame(w);
#else
	w->hextris.mono = (DefaultDepthOfScreen(XtScreen(w)) < 2 ||
		w->hextris.mono);
	for (poly = 0; poly < MAX_START_POLYHEXES; poly++) {
		w->hextris.thingGC[poly] = NULL;
		w->hextris.thingBrightGC[poly] = NULL;
		w->hextris.thingDarkGC[poly] = NULL;
		w->hextris.thingBrighterGC[poly] = NULL;
		w->hextris.thingDarkerGC[poly] = NULL;
	}
	w->hextris.borderGC = NULL;
	w->hextris.frameGC = NULL;
	w->hextris.inverseGC = NULL;
	w->hextris.xorGC = NULL;
#endif
	w->hextris.field = NULL;
	w->hextris.started = False;
	w->hextris.quit = False;
	w->hextris.completedRows = 0;
	w->hextris.newRows = 0;
	w->hextris.fall = False;
	w->hextris.focus = False;
	w->hextris.next.units = w->hextris.units;
	w->hextris.next.corners = w->hextris.corners;
	w->hextris.current.corners = w->hextris.corners;
	w->hextris.next.mixed = w->hextris.mixed;
	w->hextris.next.bonus = False;
	w->hextris.startLevel = w->hextris.level;
	w->hextris.nextThing.units = 0;
	w->hextris.nextThing.polyhexNumber = 0;
	w->hextris.nextThing.xpos = 0;
	w->hextris.nextThing.ypos = 0;
	w->hextris.nextThing.size = 0;
	w->hextris.nextThing.colorNumber = 0;
	w->hextris.nextThing.direction = 0;
	w->hextris.nextThing.bonus = 0;
	w->hextris.nextThing.randomRotation = 0;
	w->hextris.nextThing.randomReflection = 0;
	w->hextris.nextThing.randomNumber = 0;
	w->hextris.gotNew = False;
	CheckThings(w);
	SizeGame(w);
#ifdef WINVER
	brush = CreateSolidBrush(w->hextris.inverseGC);
	SETBACK(w->core.hWnd, brush);
	(void) SRAND(time(NULL));
#else
	(void) SRAND(getpid());
	SetAllColors(w);
#endif
	w->hextris.invert = (NRAND(2) % 2) ? True : False;
	corners = (w->hextris.next.corners) ? 1 : 0;
	mixed = (w->hextris.next.mixed) ? 1 : 0;
	readThings();
	turnStyle = tris[w->hextris.next.units -
		MIN_UNITS].mode[corners].turnStyle;
	/* use intCat */
	if (turnStyle < 0 || turnStyle >= MAX_TURNSTYLE_TYPES) {
		(void) fprintf(stderr, "Hextris: corrupted polyhex file\n");
		(void) fprintf(stderr, "\tturnStyle = %d.\n", turnStyle);
		exit(1);
	}
	w->hextris.numberPolyhexes = tris[w->hextris.next.units -
		MIN_UNITS].number[mixed][turnStyle];
	if (w->hextris.numberPolyhexes <= 0 ||
			w->hextris.numberPolyhexes > MAX_START_POLYHEXES) {
		(void) fprintf(stderr, "Hextris: corrupted polyhex file\n");
		(void) fprintf(stderr, "\tnumberPolyhexes = %d.\n",
			w->hextris.numberPolyhexes);
		exit(1);
	}
	ClearThings(w, False);
	newThing(w);
	newThing(w);
}

#ifndef WINVER
static
#endif
void
ExposeGame(
#ifdef WINVER
HextrisWidget w
#else
Widget renew, XEvent *event, Region region
#endif
)
{
#ifndef WINVER
	HextrisWidget w = (HextrisWidget) renew;

	if (!w->core.visible)
		return;
#endif
	EraseFrame(w);
#if 0
	ResizeThings(w);
#endif
	DrawFrame(w, w->hextris.focus);
	if (w->hextris.grid)
		drawGrid(w);
	drawField(w);
	drawThing(w);
	drawNext(w);
}

#ifndef WINVER
static
#endif
void
HideGame(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (w->hextris.sound)
		w->hextris.sound = False;
	w->hextris.paused = True;
	setGame(w, GAME_HIDE);
}

#ifndef WINVER
static
#endif
void
ClearGame(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	ClearThings(w, True);
}

#ifndef WINVER
static
#endif
void
EnterGame(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	w->hextris.focus = True;
	w->hextris.paused = False;
	DrawFrame(w, w->hextris.focus);
}

#ifndef WINVER
static
#endif
void
LeaveGame(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	/* if (w->hextris.firstFall) */
	w->hextris.focus = False;
	w->hextris.paused = True;
	DrawFrame(w, w->hextris.focus);
}

#ifndef WINVER
static void
ReflectGame(HextrisWidget w, XEvent *event, char **args, int nArgs)
{
	MoveGameInput(w, REFLECT);
}

static void
RotateGame(HextrisWidget w, XEvent *event, char **args, int nArgs)
{
	MoveGameInput(w, ROTATE);
}

static void
MoveGameLeft(HextrisWidget w, XEvent *event, char **args, int nArgs)
{
	MoveGameInput(w, MOVE_LEFT);
}

static void
MoveGameRight(HextrisWidget w, XEvent *event, char **args, int nArgs)
{
	MoveGameInput(w, MOVE_RIGHT);
}

static void
DropGame(HextrisWidget w, XEvent *event, char **args, int nArgs)
{
	MoveGameInput(w, DROP);
}

#endif

#ifdef WINVER
Boolean
#else
static void
#endif
IncrementGameLevel(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->hextris.paused && w->hextris.level < MAX_LEVELS) {
		w->hextris.level++;
		w->hextris.delay = delay[w->hextris.level - 1];
		w->hextris.newRows = 0;
		drawStatus(w);
#ifdef WINVER
		return True;
#endif
	}
#ifdef WINVER
		return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
DecrementGameLevel(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
#ifdef CHEAT
	if (!w->hextris.paused && w->hextris.level > MIN_LEVELS) {
		w->hextris.level--;
		w->hextris.delay = delay[w->hextris.level - 1];
		w->hextris.newRows = 0;
		drawStatus(w);
#ifdef WINVER
		return True;
#endif
	}
#endif
#ifdef WINVER
		return False;
#endif
}

#ifndef WINVER
static
#endif
void
PauseGame(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (w->hextris.sound)
		w->hextris.sound = False;
	w->hextris.paused = !w->hextris.paused;
#if 0
	if (w->hextris.paused) {
		banner(MSG_PAUSED);
		(void) XIconifyWindow(display, mainWin, screen_num);
	} else {
		drawNext(w);
	}
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
IncrementGameUnits(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->hextris.paused && LEGAL(w->hextris.next.units + 1,
			w->hextris.next.corners, w->hextris.next.mixed,
			False)) {
		setGame(w, GAME_INCUNITS);
#ifdef WINVER
		w->hextris.next.units = w->hextris.units;
		w->hextris.next.bonus = False;
		w->hextris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
DecrementGameUnits(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->hextris.paused && LEGAL(w->hextris.next.units - 1,
			w->hextris.next.corners, w->hextris.next.mixed,
			False)) {
		setGame(w, GAME_DECUNITS);
#ifdef WINVER
		w->hextris.next.units = w->hextris.units;
		w->hextris.next.bonus = False;
		w->hextris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
IncrementGameColumns(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->hextris.paused && w->hextris.columns < MAX_COLUMNS) {
#ifdef WINVER
		w->hextris.paused = True;
		w->hextris.started = False;
		EraseFrame(w);
		w->hextris.columns++;
		SizeGame(w);
#if 0
		ResizeThings(w);
#endif
		newThing(w);
		newThing(w);
		ClearThings(w, True);
		DrawFrame(w, w->hextris.focus);
		if (w->hextris.grid)
			drawGrid(w);
		return True;
#else
		setGame(w, GAME_INCCOLUMNS);
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
DecrementGameColumns(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->hextris.paused && w->hextris.columns > MIN_COLUMNS) {
#ifdef WINVER
		w->hextris.paused = True;
		w->hextris.started = False;
		EraseFrame(w);
		w->hextris.columns--;
		SizeGame(w);
#if 0
		ResizeThings(w);
#endif
		newThing(w);
		newThing(w);
		ClearThings(w, True);
		DrawFrame(w, w->hextris.focus);
		if (w->hextris.grid)
			drawGrid(w);
		return True;
#else
		setGame(w, GAME_DECCOLUMNS);
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifndef WINVER
static
#endif
void
FixedFillGame(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FIXEDFILL);
}

#ifndef WINVER
static
#endif
void
SuddenAppearGame(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_SUDDENAPPEAR);
#ifdef WINVER
	redoNext(w);
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
CornersGame(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->hextris.paused && LEGAL(w->hextris.next.units,
			!w->hextris.next.corners, w->hextris.next.mixed,
			False)) {
		setGame(w, GAME_CORNERS);
#ifdef WINVER
		w->hextris.next.corners = w->hextris.corners;
		w->hextris.next.bonus = False;
		w->hextris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
MixedGame(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->hextris.paused && LEGAL(w->hextris.next.units,
			w->hextris.next.corners, !w->hextris.next.mixed,
			False)) {
		setGame(w, GAME_MIXED);
#ifdef WINVER
		w->hextris.next.mixed = w->hextris.mixed;
		w->hextris.next.bonus = False;
		w->hextris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifndef WINVER
static
#endif
void
BonusGame(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_BONUS);
#ifdef WINVER
	w->hextris.next.bonus = False;
	w->hextris.newRows = 0;
	redoNext(w);
	drawNext(w);
#endif
}

#ifndef WINVER
static
#endif
void
CwGame(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_CW);
}

#ifndef WINVER
static
#endif
void
PlainPieceGame(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_PLAINPIECE);
#ifdef WINVER
	drawField(w);
#endif
}

#ifndef WINVER
static
#endif
void
ShowNextGame(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->hextris.paused) {
		setGame(w, GAME_SHOWNEXT);
	}
}

#ifndef WINVER
static
#endif
void
GridGame(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->hextris.paused) {
		setGame(w, GAME_GRID);
#ifndef WINVER
		EraseFrame(w);
		DrawFrame(w, w->hextris.focus);
		if (w->hextris.grid)
			drawGrid(w);
		drawField(w);
#endif
	}
}

#ifndef WINVER
static
#endif
void
SoundGame(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_SOUND);
}

#ifndef WINVER
static
#endif
void
Fill0Game(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL0);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill1Game(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL1);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill2Game(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL2);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill3Game(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL3);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill4Game(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL4);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill5Game(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL5);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill6Game(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL6);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill7Game(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL7);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill8Game(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL8);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill9Game(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL9);
#ifdef WINVER
	ClearThings(w, True);
#endif
}

#ifndef WINVER
static
#endif
void
Fill10Game(HextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, GAME_FILL10);
#ifdef WINVER
	ClearThings(w, True);
#endif
}
